# Copyright 2020 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Utilities and conventions for GitLab tags.
"""
GITLAB_TAG_REF_PREFIX = b'refs/tags/'


def gitlab_tag_ref(gitlab_tag):
    return GITLAB_TAG_REF_PREFIX + gitlab_tag



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Integration tests for the Git subsystem

In this case, "integration" means that we're testing the main methods,
using true Git repos, and the whole Hook logic. We may also need to create
complex transactions, with several commits, rebases, phase changes etc.
"""
from __future__ import absolute_import

from dulwich.protocol import ZERO_SHA
import pytest
import re
import subprocess

from heptapod.gitlab import hooks
from heptapod.gitlab.prune_reasons import (
    HeadPruneReason,
    BookmarkRemoved,
    BranchClosed,
    TopicPublished,
    WildHeadResolved,
)
from mercurial_testhelpers.repo_wrapper import NULL_ID
from mercurial_testhelpers.util import as_bytes
from heptapod.testhelpers import (
    RepoWrapper,
)
from hgext3rd.heptapod.branch import (
    get_default_gitlab_branch,
    read_gitlab_branches,
)
from mercurial import (
    encoding,
    error,
    pycompat,
    scmutil,
    ui as uimod,
)
from ...git import (
    GitRefChange,
    HeptapodGitHandler,
)
from ..utils import common_config, HEPTAPOD_REQUIRED_HGRC


parametrize = pytest.mark.parametrize


def patch_gitlab_hooks(monkeypatch, records, action=None):

    def call(self, changes):
        records.append((self.name, changes))
        if action is not None:
            return action(self.name, changes)
        else:
            return 0, ("hook %r ok" % self.name).encode(), 'no error'

    def init(self, repo, encoding='utf-8'):
        self.repo = repo
        self.encoding = encoding

    monkeypatch.setattr(hooks.Hook, '__init__', init)
    monkeypatch.setattr(hooks.PreReceive, '__call__', call)
    monkeypatch.setattr(hooks.PostReceive, '__call__', call)


# not covering a log interceptor for tests is not a break of coverage
# perhaps this utility should move to pytest-mercurial
def patch_ui_warnings(monkeypatch, records):  # pragma no cover

    def warn(self, *a):
        records.append(a)

    monkeypatch.setattr(uimod.ui, 'warn', warn)


def extract_git_branch_titles(branches):
    return {ref: info['title'] for ref, info in branches.items()}


class GitRepo(object):

    def __init__(self, path):
        self.path = path

    @classmethod
    def init(cls, path):
        subprocess.call(('git', 'init', '--bare', str(path)))
        return cls(path)

    def branch_titles(self):
        return extract_git_branch_titles(self.branches())

    def branches(self):
        out = subprocess.check_output(('git', 'branch', '-v', '--no-abbrev'),
                                      cwd=str(self.path))
        split_lines = (l.lstrip(b'*').strip().split(None, 2)
                       for l in out.splitlines())
        return {sp[0]: dict(sha=sp[1], title=sp[2]) for sp in split_lines}

    def tags(self):
        out = subprocess.check_output(('git', 'tag'), cwd=str(self.path))
        return set(l.strip() for l in out.splitlines())

    def commit_hash_title(self, revspec):
        out = subprocess.check_output(
            ('git', 'log', '-n1', revspec, r'--pretty=format:%H|%s'),
            cwd=str(self.path))
        return out.strip().split(b'|')

    def get_symref(self, name):
        return self.path.join(name).read().strip().split(':', 1)[1].strip()

    def set_symref(self, name, target_ref):
        self.path.join(name).write('ref: %s\n' % target_ref)

    def set_branch(self, name, sha):
        sha = pycompat.sysstr(sha)
        self.path.join('refs', 'heads', name).ensure().write(sha + '\n')

    def get_branch_sha(self, name):
        if isinstance(name, bytes):
            name = name.decode()
        return self.path.join('refs', 'heads', name).read().strip()


def make_empty_repo(path):
    config = common_config()
    config['extensions']['hggit'] = ''
    config['phases'] = dict(publish=False)

    return RepoWrapper.init(path, config=config)


def make_main_repo(path):
    """Make repo with 2 public revs; return wrapper, ctx of rev 0

    The returned ctx is for the first changeset because we'll use it as
    a branching point, hence more often than the second.
    """
    wrapper = make_empty_repo(path)
    ctx = wrapper.write_commit('foo', content='foo0', message="default0",
                               return_ctx=True)
    wrapper.write_commit('foo', content='foo1', message="default1")
    wrapper.set_phase('public', ['.'])
    return wrapper, ctx


def set_allow_bookmarks(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.bookmarks-on-named-branches', value)


def set_allow_multiple_heads(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(b'heptapod', b'allow-multiple-heads', value)


def set_prune_closed_branches(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-newly-closed-branches', value)


def activate_mirror(repo_wrapper):
    """Activate the mirrorring hook for given repo.

    Using the hook is a simple way to get in-transaction, while still within
    a repo fully controlled by these tests (configuration, notably).
    """
    repo_wrapper.repo.ui.setconfig(
        b'hooks', b'pretxnclose.testcase',
        b'python:heptapod.hooks.gitlab_mirror.mirror')


def test_basic(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    repo, base_ctx = make_main_repo(repo_path)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    repo.command('gitlab-mirror')

    assert git_repo.branch_titles() == {b'branch/default': b'default1'}
    sha = git_repo.branches()[b'branch/default']['sha']
    assert notifs == [
        ('pre-receive', ({}, {b'refs/heads/branch/default': (ZERO_SHA, sha)})),
        ('post-receive', ({},
                          {b'refs/heads/branch/default': (ZERO_SHA, sha)})),
    ]

    assert read_gitlab_branches(repo.repo) == {
        b'branch/default': repo.repo[b'.'].hex()
    }


def test_alternate_encoding(tmpdir, monkeypatch):
    """Round trip with branch name not in UTF-8.

    It's not clear whether this is valid on the GitLab side, but it should
    be transparent in here: what comes in comes out.
    """
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    wrapper, base_ctx = make_main_repo(repo_path)
    wrapper.command('gitlab-mirror')
    notifs[:] = []

    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    # Even though it's been initialized from the `HGENCODING` environment
    # variable, the encoding is a global.
    monkeypatch.setattr(encoding, 'encoding', b'latin-1')

    ctx = wrapper.write_commit('encoded', branch=b'pr\xe9parations')
    wrapper.command('gitlab-mirror')

    # not using existing helpers to avoid making the test tautological
    git_branch = b'branch/pr\xe9parations'
    git_ref = b'refs/heads/' + git_branch

    # both in Git repo and in notifs, we have the exact same bytes as provided
    assert set(git_repo.branch_titles()) == {b'branch/default', git_branch}
    sha = git_repo.branches()[git_branch]['sha']
    assert notifs == [
        ('pre-receive', ({}, {git_ref: (ZERO_SHA, sha)})),
        ('post-receive', ({}, {git_ref: (ZERO_SHA, sha)})),
    ]
    assert read_gitlab_branches(wrapper.repo)[git_branch] == ctx.hex()


def test_wiki(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    wrapper, base_ctx = make_main_repo(repo_path)
    wrapper.repo.ui.environ[b'GL_REPOSITORY'] = b'wiki-251'
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.command('gitlab-mirror')

    # for wikis, we duplicate `branch/default` as `master`
    assert git_repo.branch_titles() == {b'branch/default': b'default1',
                                        b'master': b'default1',
                                        }
    sha = git_repo.branches()[b'branch/default']['sha']
    changes = {}, {b'refs/heads/branch/default': (ZERO_SHA, sha),
                   b'refs/heads/master': (ZERO_SHA, sha),
                   }
    assert notifs == [
        ('pre-receive', changes),
        ('post-receive', changes),
    ]


def test_tags(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    repo, base_ctx = make_main_repo(repo_path)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    # Creation
    repo.command('tag', b'v1.2.3', rev=base_ctx.hex())
    repo.command('gitlab-mirror')

    branches = git_repo.branches()
    assert list(branches) == [b'branch/default']
    branch = branches[b'branch/default']

    assert b'v1.2.3' in branch['title']
    assert git_repo.tags() == {b'v1.2.3', }
    tagged_git_sha_0, tagged_git_title = git_repo.commit_hash_title('v1.2.3')
    assert tagged_git_title == b'default0'

    first_default_branch_sha = branch['sha']
    changes = {}, {
        b'refs/heads/branch/default': (ZERO_SHA, first_default_branch_sha),
        b'refs/tags/v1.2.3': (ZERO_SHA, tagged_git_sha_0),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    del notifs[:]

    # Modification
    repo.command('tag', b'v1.2.3', rev=b'1', force=True)
    repo.command('gitlab-mirror')

    assert git_repo.tags() == {b'v1.2.3', }
    tagged_git_sha_1, tagged_git_title = git_repo.commit_hash_title('v1.2.3')
    assert tagged_git_title == b'default1'

    new_default_branch_sha = git_repo.branches()[b'branch/default']['sha']
    changes = {}, {
        b'refs/heads/branch/default': (first_default_branch_sha,
                                       new_default_branch_sha),
        b'refs/tags/v1.2.3': (tagged_git_sha_0, tagged_git_sha_1),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]

    # Removal not supported in Heptapod 0.8. TODO later


def test_tags_obsolete(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)

    # we'll need to perform a pull in order to amend a tagged changeset
    # and rebase the tagging changeset in a single transaction.
    src_path = tmpdir / 'src.hg'
    src = RepoWrapper.init(src_path, config=common_config())

    dest = make_empty_repo(tmpdir / 'dest.hg')
    activate_mirror(dest)

    def dest_pull():
        dest.command('pull', source=as_bytes(src_path), force=True)

    # Creation
    src.commit_file('foo')

    # no problem on receiving side with obsolescence of changesets
    # that it never received
    src.commit_file('bar')
    src.command("amend", message=b'amend before exchange')
    dest_pull()

    src.command('tag', b'v1.2.3')
    tag_ctx = scmutil.revsingle(src.repo, b'.')
    dest_pull()

    tagged = scmutil.revsingle(src.repo, b'v1.2.3')
    src.update_bin(tagged.node())
    src_path.join("foo", "amending")
    src.command('amend', message=b'amend')
    src.command('rebase', rev=[tag_ctx.hex()])

    with pytest.raises(error.Abort) as exc_info:
        dest_pull()
    assert re.search(br'tag.*v1\.2\.3.*obsolete', exc_info.value.args[0])


def test_tip_obsolete(tmpdir, monkeypatch):
    """Since 'tip' is a tag, we don't to refuse making it obsolete."""
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)

    # we'll need to perform a pull in order to amend a tagged changeset
    # and rebase the tagging changeset in a single transaction.
    repo_path = tmpdir / 'repo.hg'
    wrapper = RepoWrapper.init(repo_path, config=common_config())
    activate_mirror(wrapper)  # we'll need the mirror in-transaction

    wrapper.commit_file('foo')

    ctx = wrapper.commit_file('foo')
    wrapper.prune(ctx.hex())  # no error

    # yet the tip of the unfiltered repo is obsolete
    # (if this change in a later Mercurial version, we may get rid of
    # this assertion)
    assert b'tip' in wrapper.repo.unfiltered()[ctx.node()].tags()


def test_share(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    main_path = tmpdir.join('main.hg')
    main_wrapper, base_ctx = make_main_repo(main_path)
    git_repo = GitRepo.init(tmpdir.join('main.git'))

    # let's start with some commits in the Git repo
    main_wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}

    # now let's make a share
    dest_wrapper = main_wrapper.share(tmpdir.join('share.hg'))
    dest_wrapper.write_commit('bar', message='other0',
                              branch='other', parent=base_ctx.node())
    dest_wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1',
                                        b'branch/other': b'other0'}


def test_bookmarks(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)

    wrapper = RepoWrapper.init(tmpdir.join('repo'), config=common_config())

    # this activates in particular the gitlab-mirror hook
    wrapper.repo.ui.readconfig(HEPTAPOD_REQUIRED_HGRC, trust=True)

    set_allow_bookmarks(wrapper, True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    base = wrapper.commit_file('foo', message="unbookmarked")
    bk1 = wrapper.commit_file('foo', message="book1")
    bk2 = wrapper.commit_file('foo', message="book2")
    wrapper.command('bookmark', b'zebook1', rev=bk1.hex())
    wrapper.command('bookmark', b'zebook2', rev=bk2.hex())
    assert git_repo.branch_titles() == {b'branch/default': b'book2',
                                        b'zebook1': b'book1',
                                        b'zebook2': b'book2'}
    assert read_gitlab_branches(wrapper.repo) == {b'branch/default': bk2.hex(),
                                                  b'zebook1': bk1.hex(),
                                                  b'zebook2': bk2.hex()}
    ctx = wrapper.commit_file('foo', message="new default head", parent=base)
    assert git_repo.branch_titles() == {b'branch/default': b'new default head',
                                        b'zebook1': b'book1',
                                        b'zebook2': b'book2'}
    assert read_gitlab_branches(wrapper.repo) == {b'branch/default': ctx.hex(),
                                                  b'zebook1': bk1.hex(),
                                                  b'zebook2': bk2.hex()}

    # and multiple heads are still refused
    with pytest.raises(error.Abort) as exc:
        wrapper.commit_file('foo', message="2nd default head", parent=base)
    assert b'multiple heads' in exc.value.args[0]


def test_bookmarks_prune(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    server_path = tmpdir.join('repo.hg')
    server, base_ctx = make_main_repo(server_path)
    repo = server.repo
    default1_ctx = repo[b'.']
    server.command('bookmark', b'zebook', rev=base_ctx.hex())
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    # not being in a transaction accepts the bookmark immediately
    server.command('gitlab-mirror')
    del notifs[:]

    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'zebook': b'default0'
    }
    book_git_sha = git_repo.branches()[b'zebook']['sha']

    activate_mirror(server)
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(base_ctx.node()) == []
    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
    }

    assert read_gitlab_branches(repo) == {
        b'branch/default': default1_ctx.hex()
    }

    changes = {b'zebook': BookmarkRemoved(book_git_sha)}, {
        b'refs/heads/zebook': (book_git_sha, ZERO_SHA)
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]


def test_bookmarks_mask_branch_prune(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    server_path = tmpdir.join('repo.hg')
    server, base_ctx = make_main_repo(server_path)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    server.command('gitlab-mirror')
    # just checking our assumptions
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}

    # we need to test the branch masking on a branch
    # that's not the GitLab default (which is protected)
    server.write_commit('foo', branch='other', message='other1')

    activate_mirror(server)
    set_allow_bookmarks(server, True)
    head = server.repo[b'tip']
    server.command('bookmark', b'zebook', rev=head.hex())
    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'zebook': b'other1'
    }
    book_git_sha = git_repo.branches()[b'zebook']['sha']

    del notifs[:]
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(head.node()) == []
    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'branch/other': b'other1',
    }

    changes = {b'zebook': BookmarkRemoved(book_git_sha)}, {
        b'refs/heads/zebook': (book_git_sha, ZERO_SHA),
        b'refs/heads/branch/other': (ZERO_SHA, book_git_sha),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]


def test_bookmarks_obsolete(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    config = common_config()
    config['extensions']['hggit'] = ''
    config['phases'] = dict(publish=False)
    wrapper = RepoWrapper.init(repo_path, config=config)
    set_allow_bookmarks(wrapper, True)

    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    # getting gitlab-mirror to work in-transaction

    draft1 = wrapper.commit_file('foo', message='Initial draft')
    wrapper.command('bookmark', b'zebook', rev=draft1.hex(), inactive=True)
    wrapper.command('gitlab-mirror')

    # just checking our assumptions.
    assert git_repo.branch_titles() == {b'zebook': b'Initial draft'}

    # let's get in-transaction
    activate_mirror(wrapper)
    draft2 = wrapper.commit_file('foo', message='amended', parent=NULL_ID)
    with pytest.raises(error.Abort) as exc_info:
        wrapper.prune(draft1.hex(), successors=[draft2.hex()])
    assert re.search(br'bookmark.*zebook.*obsolete', exc_info.value.args[0])


def test_bookmarks_dont_mask_default_branch(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    server_path = tmpdir.join('repo.hg')
    server, base_ctx = make_main_repo(server_path)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    server.command('gitlab-mirror')
    # just checking our assumptions
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}

    activate_mirror(server)
    set_allow_bookmarks(server, True)
    head = server.repo[b'tip']
    git_sha = git_repo.get_branch_sha('branch/default').encode()

    del notifs[:]
    server.command('bookmark', b'zebook', rev=head.hex())

    # the default branch is not pruned
    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'zebook': b'default1'
    }

    changes = {b'refs/heads/zebook': (ZERO_SHA, git_sha)}
    assert notifs == [('pre-receive', ({}, changes)),
                      ('post-receive', ({}, changes)),
                      ]

    del notifs[:]
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(head.node()) == []
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}
    changes = {b'refs/heads/zebook': (git_sha, ZERO_SHA)}
    prunes = {b'zebook': BookmarkRemoved(git_sha)}

    assert notifs == [('pre-receive', (prunes, changes)),
                      ('post-receive', (prunes, changes)),
                      ]


@parametrize('branch_name', ('default', 'other'))
def test_change_gitlab_default_branch(tmpdir, monkeypatch, branch_name):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    config = common_config()
    config['extensions']['hggit'] = ''
    config['phases'] = dict(publish=False)

    wrapper = RepoWrapper.init(tmpdir.join('repo.hg'), config=config)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('foo', message="other0",
                         branch=branch_name,
                         topic='initial')
    wrapper.command('gitlab-mirror')

    def assert_default_gitlab_branch(gl_branch):
        assert git_repo.get_symref('HEAD') == 'refs/heads/' + gl_branch
        gl_branch_bytes = pycompat.sysbytes(gl_branch)
        assert get_default_gitlab_branch(wrapper.repo) == gl_branch_bytes

    assert_default_gitlab_branch('topic/%s/initial' % branch_name)

    wrapper.set_phase('public', ['.'])
    wrapper.command('gitlab-mirror')

    assert_default_gitlab_branch('branch/' + branch_name)


def test_change_gitlab_default_branch_nothing_new(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    config = common_config()
    config['extensions']['hggit'] = ''
    config['phases'] = dict(publish=False)

    wrapper = RepoWrapper.init(tmpdir.join('repo.hg'), config=config)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('foo', message="other0",
                         branch='default',
                         topic='initial')
    wrapper.command('gitlab-mirror')

    def assert_default_gitlab_branch(gl_branch):
        assert git_repo.get_symref('HEAD') == 'refs/heads/' + gl_branch
        gl_branch_bytes = pycompat.sysbytes(gl_branch)
        assert get_default_gitlab_branch(wrapper.repo) == gl_branch_bytes

    assert_default_gitlab_branch('topic/default/initial')

    wrapper.write_commit('foo', message="same gitlab branch")
    wrapper.command('gitlab-mirror')

    assert_default_gitlab_branch('topic/default/initial')


def test_closed_branch(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    wrapper.write_commit('foo', message="other0",
                         branch='other',
                         parent=base_ctx.node(),
                         return_ctx=True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1',
                                        b'branch/other': b'other0'}
    to_close_sha = git_repo.branches()[b'branch/other']['sha']
    set_prune_closed_branches(wrapper, True)
    activate_mirror(wrapper)
    del notifs[:]

    wrapper.command('commit', message=b"closing other",
                    close_branch=True)
    # we'll let GitLab do the pruning so that it can use the closing
    # sha for Merge Request detection.
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    closing_sha = handler.map_git_get(wrapper.repo[b'tip'].hex())
    prune_reason = {b'branch/other':
                    BranchClosed([(closing_sha, [to_close_sha])])}
    changes = {b'refs/heads/branch/other': (to_close_sha, ZERO_SHA)}

    assert notifs == [('pre-receive', (prune_reason, changes)),
                      ('post-receive', (prune_reason, changes))]


def test_previously_closed_branch_not_pruned(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    wrapper.write_commit('foo', message="other0",
                         branch='other',
                         parent=base_ctx.node(),
                         return_ctx=True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    # let's prepare a closed branch that hasn't been prune
    set_prune_closed_branches(wrapper, False)
    wrapper.command('commit', message=b"closing other", close_branch=True)
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1',
                                        b'branch/other': b'closing other'}

    # subsequent calls won't prune it...
    set_prune_closed_branches(wrapper, True)
    wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-previously-closed-branches', False)
    wrapper.command('gitlab-mirror')

    assert git_repo.branch_titles()[b'branch/other'] == b'closing other'

    # until we flip the right switch
    wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-previously-closed-branches', True)
    wrapper.command('gitlab-mirror')

    assert b'branch/other' not in git_repo.branch_titles()


def test_closed_branch_not_in_git(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    wrapper.write_commit('foo', message="other0",
                         branch='other',
                         parent=base_ctx.node(),
                         return_ctx=True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    # the `other` branch being mirrored as already closed, will trigger
    # a prune request that should be ignored in order not to fail
    wrapper.command('commit', message=b"closing other", close_branch=True)
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}


def test_closed_default_branch(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}
    git_repo.set_symref('HEAD', 'refs/heads/branch/default')

    set_prune_closed_branches(wrapper, True)
    wrapper.command('commit', message=b"closing default!", close_branch=True)

    # On native repos, this is a user error
    wrapper.repo.ui.setconfig(b'heptapod', b'native', True)

    del notifs[:]
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    assert re.search(br'prune.*default branch', exc_info.value.args[0])

    assert not notifs
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}

    # On non-native repos, this is ignored
    wrapper.repo.ui.setconfig(b'heptapod', b'native', False)
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'default1'}


def test_multiple_heads_merge(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)

    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    set_allow_multiple_heads(wrapper, True)
    first = wrapper.repo[b'tip']
    second = wrapper.write_commit('bar', message="second head",
                                  branch='default',
                                  parent=base_ctx.node(),
                                  return_ctx=True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {
        b'wild/' + first.hex(): b'default1',
        b'wild/' + second.hex(): b'second head',
        # the most recently added revision always wins
        b'branch/default': b'second head',
    }
    git_branches = git_repo.branches()
    first_git_sha = git_branches[b'wild/' + first.hex()]['sha']
    second_git_sha = git_branches[b'wild/' + second.hex()]['sha']
    assert read_gitlab_branches(wrapper.repo) == {
        b'wild/' + first.hex(): first.hex(),
        b'wild/' + second.hex(): second.hex(),
        b'branch/default': second.hex(),
    }
    wrapper.command('merge')
    wrapper.command('commit', message=b'merging heads')
    merge_ctx = wrapper.repo[b'.']
    del notifs[:]
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {b'branch/default': b'merging heads'}
    assert read_gitlab_branches(wrapper.repo) == {
        b'branch/default': merge_ctx.hex(),
    }

    prune_reasons = {
        b'wild/' + first.hex(): WildHeadResolved(first_git_sha),
        b'wild/' + second.hex(): WildHeadResolved(second_git_sha),
    }
    changes = {
        b'refs/heads/wild/' + first.hex(): (first_git_sha, ZERO_SHA),
        b'refs/heads/wild/' + second.hex(): (second_git_sha, ZERO_SHA),
        b'refs/heads/branch/default': (
            second_git_sha,
            git_repo.branches()[b'branch/default']['sha'])
    }

    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_push_multiple_heads_switch_branch(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    set_allow_multiple_heads(wrapper, True)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    # second head on default branch (easy to merge for later test)
    wrapper.write_commit('bar', message="default head 2",
                         parent=base_ctx.node())
    default_heads_titles = {b'default1', b'default head 2'}

    wrapper.command('gitlab-mirror')

    branch_titles = git_repo.branch_titles()
    assert set(branch_titles.values()) == default_heads_titles
    assert len(branch_titles) == 3

    wrapper.write_commit('foo', message="other", branch='other')
    wrapper.command('gitlab-mirror')
    branch_titles = git_repo.branch_titles()
    assert set(branch_titles.values()) == default_heads_titles | {b'other', }
    assert len(branch_titles) == 4

    # now let's add a topic on top of one of those wild 'default' heads
    wrapper.write_commit('foo', message="on topic",
                         topic='zetop',
                         parent=base_ctx.node())

    wrapper.command('gitlab-mirror')
    branch_titles = git_repo.branch_titles()
    assert set(b for b in branch_titles if not b.startswith(b'wild/')) == {
        b'branch/default', b'branch/other', b'topic/default/zetop'}

    assert set(title
               for name, title in branch_titles.items()
               if name.startswith(b'wild/')) == default_heads_titles
    assert branch_titles[b'branch/default'] in default_heads_titles
    assert branch_titles[b'branch/other'] == b'other'
    assert branch_titles[b'topic/default/zetop'] == b'on topic'


def test_push_multiple_heads_refuse(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))

    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('bar', message="default head 2",
                         parent=base_ctx.node())

    # default behaviour:
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    # with explicit config:
    set_allow_multiple_heads(wrapper, False)
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    assert b'multiple heads' in exc_info.value.args[0]
    assert git_repo.branch_titles() == {}


def test_topic_pruned(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, _ = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('foo', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    del notifs[:]

    topic_gl_branch = b'topic/default/zzetop'
    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        topic_gl_branch: b'in topic'}

    wrapper.prune(b'zzetop')
    wrapper.command('gitlab-mirror')

    topic_ref = b'refs/heads/topic/default/zzetop'
    topic_change = {topic_ref: (
        branches_before[b'topic/default/zzetop']['sha'],
        ZERO_SHA,
    )}
    prune_reason = {topic_gl_branch: HeadPruneReason()}

    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
    }
    assert notifs == [('pre-receive', (prune_reason, topic_change)),
                      ('post-receive', (prune_reason, topic_change))]


def test_topic_amended(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    repo_path = tmpdir.join('repo.hg')
    wrapper, _ = make_main_repo(repo_path)
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('foo', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    del notifs[:]

    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        b'topic/default/zzetop': b'in topic'}

    repo_path.join('foo').write('amend1')
    wrapper.command('amend', message=b'amend1')

    wrapper.command('gitlab-mirror')

    branches = git_repo.branches()
    assert extract_git_branch_titles(branches) == {
        b'branch/default': b'default1',
        b'topic/default/zzetop': b'amend1',
    }

    topic_change = {b'refs/heads/topic/default/zzetop': (
        branches_before[b'topic/default/zzetop']['sha'],
        branches[b'topic/default/zzetop']['sha'],
    )}

    assert notifs == [('pre-receive', ({}, topic_change)),
                      ('post-receive', ({}, topic_change))]


def test_topic_ff_publish(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, _ = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.command('topics', b'zzetop')
    wrapper.write_commit('foo', message='in ff topic')
    wrapper.command('gitlab-mirror')
    del notifs[:]

    topic_gl_branch = b'topic/default/zzetop'
    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        topic_gl_branch: b'in ff topic'}

    wrapper.set_phase('public', ['zzetop'])
    wrapper.command('gitlab-mirror')

    assert git_repo.branch_titles() == {b'branch/default': b'in ff topic'}

    topic_ref = b'refs/heads/topic/default/zzetop'
    topic_before_sha = branches_before[b'topic/default/zzetop']['sha']

    prune_reasons = {topic_gl_branch: TopicPublished(topic_before_sha)}
    changes = {
        b'refs/heads/branch/default': (
            branches_before[b'branch/default']['sha'],
            topic_before_sha,
        ),
        topic_ref: (
            branches_before[b'topic/default/zzetop']['sha'],
            ZERO_SHA,
        )
    }

    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_topic_rebase_publish(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, _ = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('zz', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    del notifs[:]
    topic_gl_branch = b'topic/default/zzetop'

    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        topic_gl_branch: b'in topic'}

    wrapper.command('rebase', rev=[b'zzetop'])
    wrapper.set_phase('public', ['zzetop'])
    wrapper.command('gitlab-mirror')

    branches_after = git_repo.branches()
    assert extract_git_branch_titles(branches_after) == {
        b'branch/default': b'in topic'}

    default_after_sha = branches_after[b'branch/default']['sha']
    prune_reasons = {topic_gl_branch: TopicPublished(default_after_sha)}
    changes = {
        b'refs/heads/branch/default': (
            branches_before[b'branch/default']['sha'],
            default_after_sha),
        b'refs/heads/topic/default/zzetop': (
            branches_before[b'topic/default/zzetop']['sha'],
            ZERO_SHA),
    }
    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_topic_git_escape(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, _ = make_main_repo(tmpdir.join('repo.hg'))
    del notifs[:]

    # Only test names that are possible with topics in the first place
    invalid_git_ref_names = [
        ".starts-with-dot",
        "..",
        "ends-with.lock",
        "ends-with-dot.",
    ]
    for name in invalid_git_ref_names:
        wrapper.write_commit('zz', message='in topic', topic=name)

    # This fails if topic names that are invalid in Git are not escaped
    wrapper.command('gitlab-mirror')


def test_topic_add_rebase_publish(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    top_ctx = wrapper.write_commit('zz', message='in topic',
                                   parent=base_ctx.node(), topic='zzetop')
    wrapper.command('gitlab-mirror')
    del notifs[:]

    topic_gl_branch = b'topic/default/zzetop'

    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        topic_gl_branch: b'in topic'}

    assert read_gitlab_branches(wrapper.repo) == {
        b'branch/default': scmutil.revsingle(wrapper.repo, b'default').hex(),
        topic_gl_branch: top_ctx.hex(),
    }

    wrapper.write_commit('zz', message='topic addition')
    wrapper.command('rebase')
    wrapper.set_phase('public', ['zzetop'])
    wrapper.command('gitlab-mirror')

    branches_after = git_repo.branches()
    assert extract_git_branch_titles(branches_after) == {
        b'branch/default': b'topic addition'}

    assert read_gitlab_branches(wrapper.repo) == {
        b'branch/default': scmutil.revsingle(wrapper.repo, b'default').hex(),
    }

    changes = {
        b'refs/heads/branch/default': (
            branches_before[b'branch/default']['sha'],
            branches_after[b'branch/default']['sha']),
        b'refs/heads/topic/default/zzetop': (
            branches_before[b'topic/default/zzetop']['sha'],
            ZERO_SHA),
    }
    prune_reasons = {
        topic_gl_branch:
            TopicPublished(branches_after[b'branch/default']['sha'])
    }
    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_topic_clear_publish(tmpdir, monkeypatch):
    """The topic head seen from Git is public and has changed topic.

    This is the test for heptapod#265
    The starting point can be considered to be corrupted: any topic change
    should have updated the Git branch for the topic. In this scenario
    the change is a removal, wich should have pruned the Git branch, but
    somehow the Git branch got updated to the new changeset, that doesn't
    have a topic.
    """
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    wrapper.write_commit('zz', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')

    topic_gl_branch = b'topic/default/zzetop'

    branches_before = git_repo.branches()
    assert extract_git_branch_titles(branches_before) == {
        b'branch/default': b'default1',
        topic_gl_branch: b'in topic'}

    wrapper.command('topics', rev=[b'.'], clear=True)
    wrapper.command('gitlab-mirror')
    wrapper.set_phase('public', ['.'])
    forced_topic_before_sha = git_repo.get_branch_sha('branch/default')
    forced_topic_before_sha_bytes = pycompat.sysbytes(forced_topic_before_sha)

    # here's the main point:
    git_repo.set_branch(pycompat.sysstr(topic_gl_branch),
                        forced_topic_before_sha)

    del notifs[:]
    # This used to raise LookupError
    wrapper.command('gitlab-mirror')

    branches_after = git_repo.branches()
    assert extract_git_branch_titles(branches_after) == {
        b'branch/default': b'in topic',
    }
    changes = (
        {topic_gl_branch: TopicPublished(forced_topic_before_sha_bytes)},
        {b'refs/heads/topic/default/zzetop': (
            branches_after[b'branch/default']['sha'],
            ZERO_SHA)},
    )
    assert notifs == [('pre-receive', changes),
                      ('post-receive', changes)]


def test_topic_branch_change(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    wrapper.write_commit('bar', message='other0',
                         parent=base_ctx.node(), branch='other')
    wrapper.write_commit('zz', message='in topic',
                         topic='zzetop')
    wrapper.command('gitlab-mirror')

    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'branch/other': b'other0',
        b'topic/other/zzetop': b'in topic'}

    wrapper.command('rebase', rev=[b'zzetop'], dest=b'default')
    wrapper.command('gitlab-mirror')
    assert git_repo.branch_titles() == {
        b'branch/default': b'default1',
        b'branch/other': b'other0',
        b'topic/default/zzetop': b'in topic',
    }


def test_analyse_vanished_topic_not_in_git(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    repo = wrapper.repo
    git_repo = GitRepo.init(tmpdir.join('repo.git'))

    wrapper.write_commit('zz', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')

    def return_unknown(handler, *a, **kw):
        handler.repo.test_patch_called = True
        return b'unknown-to-git'

    # let's make the published topic analysis return something unknown to Git
    monkeypatch.setattr(HeptapodGitHandler, 'published_topic_latest_hg_sha',
                        return_unknown)

    wrapper.set_phase('public', ['zzetop'])
    # Calling export directly because through the hook, such
    # monkey patching often does not work.
    HeptapodGitHandler(repo, repo.ui).export_commits()

    # we didn't fail and have been conservative with the topic GitLab branch
    assert git_repo.branch_titles() == {
        b'branch/default': b'in topic',
        b'topic/default/zzetop': b'in topic'
    }

    # note that coverage would notice if the test harness didn't work
    assert wrapper.repo.test_patch_called


def test_heptapod_notify_gitlab(tmpdir, monkeypatch):
    wrapper = RepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)

    notifs = []

    def trigger_failures(name, changes):
        if b'exc' in changes[1]:
            raise RuntimeError('trigger_failures: ' + name)
        if b'code_one' in changes[1]:
            return 1, b'', 'hook refusal'
        return 0, b'Consider a MR for ' + name.encode(), ''

    patch_gitlab_hooks(monkeypatch, notifs, action=trigger_failures)

    # minimal valid change
    change = GitRefChange(b'master', b'before', b'after')
    hook = hooks.PreReceive(wrapper.repo)
    handler.heptapod_notify_gitlab(hook, [], {b'master': change},
                                   allow_error=False)
    assert notifs == [('pre-receive',
                       ([], {b'master': (b'before', b'after')})
                       )]
    del notifs[:]

    with pytest.raises(error.Abort) as exc_info:
        handler.heptapod_notify_gitlab(hook, [], {b'code_one': change},
                                       allow_error=False)
    assert b"hook refusal" in exc_info.value.args[0]

    with pytest.raises(RuntimeError) as exc_info:
        handler.heptapod_notify_gitlab(hook, [], {b'exc': change},
                                       allow_error=False)
    assert "trigger_failures" in exc_info.value.args[0]

    # case where exception is triggered yet is accepted
    errors = []

    def record_ui_error(*args):
        errors.append(args)
    monkeypatch.setattr(handler.repo.ui, 'error', record_ui_error)
    handler.heptapod_notify_gitlab(hook, [], {b'exc': change},
                                   allow_error=True)
    assert errors[0][0].splitlines()[:2] == [
        (b"GitLab update error (GitLab 'pre-receive' hook): "
         + pycompat.sysbytes(
             # this repr() differs between py2 and py3
             repr(RuntimeError('trigger_failures: pre-receive')))),
        b"Traceback (most recent call last):"
    ]


def test_heptapod_notify_gitlab_native(tmpdir, monkeypatch):
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)

    # setup a repo pair, with two commits
    wrapper, base_ctx = make_main_repo(tmpdir.join('repo.hg'))
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    topic_ctx = wrapper.write_commit('zz', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    del notifs[:]  # start over on notifs for actual testing

    gl_branch_default = b'branch/default'
    default_hg_sha = wrapper.repo[1].hex()
    default_git_sha = git_repo.get_branch_sha(gl_branch_default).encode()

    gl_branch_topic = b'topic/default/zzetop'
    topic_hg_sha = topic_ctx.hex()
    topic_git_sha = git_repo.get_branch_sha(gl_branch_topic).encode()

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)

    hook1 = hooks.PreReceive(wrapper.repo)
    hook2 = hooks.PostReceive(wrapper.repo)

    wrapper.repo.ui.setconfig(b'heptapod', b'native', b'yes')

    # branch creation, interesting because ZERO_SHA isn't in the mappings
    change = GitRefChange(b'something', ZERO_SHA, topic_git_sha)
    handler.heptapod_notify_gitlab(hook1, [], {b'something': change})
    assert notifs == [
        ('pre-receive', ([], {b'something': (ZERO_SHA, topic_hg_sha)}))
    ]

    # branch change
    # second hook call to check for side effects in pre/post-receive succession
    change = GitRefChange(b'branch/default', default_git_sha, topic_git_sha)
    del notifs[:]
    handler.heptapod_notify_gitlab(hook1, [], {b'branch/default': change})
    assert notifs == [
        ('pre-receive', ([],
                         {b'branch/default': (default_hg_sha, topic_hg_sha)}))
    ]
    del notifs[:]
    handler.heptapod_notify_gitlab(hook2, [], {b'branch/default': change})
    assert notifs == [
        ('post-receive', ([],
                          {b'branch/default': (default_hg_sha, topic_hg_sha)}))
    ]

    # Topic published (making this realistic, but consistency shouldn't matter)
    # this tests most of the PruneReasons. If a new one not subclassing
    # HeadPruneReasonWithSha would appear, coverage drop would warn us.
    del notifs[:]
    prune_change = GitRefChange(gl_branch_topic, topic_git_sha, ZERO_SHA)
    prune_reasons = {gl_branch_topic: TopicPublished(topic_git_sha)}
    changes = {gl_branch_default: change, gl_branch_topic: prune_change}
    expected = ({gl_branch_topic: TopicPublished(topic_hg_sha)},
                {gl_branch_default: (default_hg_sha, topic_hg_sha),
                 gl_branch_topic: (topic_hg_sha, ZERO_SHA),
                 })
    handler.heptapod_notify_gitlab(hook1, prune_reasons, changes)
    assert notifs[-1] == ('pre-receive', expected)
    handler.heptapod_notify_gitlab(hook2, prune_reasons, changes)
    assert notifs[-1] == ('post-receive', expected)

    # BranchClosed.
    # This example is less realistic, we pretend the topic changeset
    # actually closes the default branch. At this low level, it doesn't matter.
    del notifs[:]
    prune_change = GitRefChange(gl_branch_default, default_git_sha, ZERO_SHA)
    changes = {gl_branch_default: prune_change}
    prune_reasons = {
        gl_branch_default: BranchClosed([(topic_git_sha, [default_git_sha])]),
    }
    expected = (
        {gl_branch_default: BranchClosed([[topic_hg_sha, [default_hg_sha]]])},
        {gl_branch_default: (default_hg_sha, ZERO_SHA)},
    )
    handler.heptapod_notify_gitlab(hook1, prune_reasons, changes)
    assert notifs[-1] == ('pre-receive', expected)
    handler.heptapod_notify_gitlab(hook2, prune_reasons, changes)
    assert notifs[-1] == ('post-receive', expected)

    # HeadPruneReason with no info to convert
    prune_reasons = {gl_branch_default: HeadPruneReason()}
    expected = (
        {gl_branch_default: HeadPruneReason()},
        {gl_branch_default: (default_hg_sha, ZERO_SHA)},
    )
    handler.heptapod_notify_gitlab(hook1, prune_reasons, changes)
    assert notifs[-1] == ('pre-receive', expected)

    # Corner case: BranchClosed without info to convert
    prune_reasons = {gl_branch_default: BranchClosed()}
    expected = (
        {gl_branch_default: BranchClosed()},
        {gl_branch_default: (default_hg_sha, ZERO_SHA)},
    )
    handler.heptapod_notify_gitlab(hook1, prune_reasons, changes)
    assert notifs[-1] == ('pre-receive', expected)


def test_subrepos(tmpdir, monkeypatch):
    """See heptapod#310 and heptapod#311."""
    notifs = []
    patch_gitlab_hooks(monkeypatch, notifs)
    main_path = tmpdir.join('repo.hg')
    GitRepo.init(tmpdir.join('repo.git'))
    wrapper, base_ctx = make_main_repo(main_path)

    nested_path = main_path.join('nested')
    nested = RepoWrapper.init(nested_path)
    nested.write_commit("bar", content="in nested")
    main_path.join('.hgsub').write("\n".join((
        "nested = ../bar",
        "[subpaths]",  # reproduction of heptapod#310
        "foo2=bar2",
    )))

    wrapper.command(b'add', subrepos=True)
    wrapper.command(b'commit', subrepos=True, message=b"invalid")
    wrapper.command('gitlab-mirror')

    wrapper.repo.ui.setconfig(b'hooks', b'precommit',
                              b'python:heptapod.hooks.subrepos.forbid_commit')
    with pytest.raises(error.Abort) as exc_info:
        wrapper.write_commit("foo")
    assert re.search(b'cannot commit.*there are subrepos',
                     exc_info.value.args[0])

    # Now let's simulate update from NULL revision, as Heptapod's
    # use of shares would do
    wrapper.command(b'update', b'0000000000000000000000', clean=True)
    nested_path.remove()

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command(b'update', b'default')
    assert b'not supported' in exc_info.value.args[0]


def test_default_git_ref(tmpdir):
    wrapper = RepoWrapper.init(tmpdir.join('repo'),
                               config=common_config())
    git_repo = GitRepo.init(tmpdir.join('repo.git'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    # default Git value
    assert handler.get_default_git_ref() == b'refs/heads/master'
    other_ref = b'refs/heads/something'
    handler.set_default_gitlab_ref(other_ref)

    assert git_repo.get_symref('HEAD') == other_ref.decode()

    # instance level cache got invalidated
    assert handler.get_default_git_ref() == other_ref



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from heptapod.testhelpers import RepoWrapper
from ..branch import (
    GITLAB_BRANCHES_FILE_NAME,
    GITLAB_BRANCHES_MISSING,
    gitlab_branches,
    read_gitlab_branches,
    write_gitlab_branches,
)


def test_read_gitlab_branches_v1(tmpdir):
    # this test depends on file layout details (so that it serves as a
    # proof of non tautology of all these tests)
    # and is specific to version 1 of the format.
    wrapper = RepoWrapper.init(tmpdir)
    sha1 = "fe54" * 10
    sha2 = "5c34" * 10
    tmpdir.join('.hg', 'store', GITLAB_BRANCHES_FILE_NAME.decode()
                ).write("\n".join((
                    "001",
                    "%s branch/some-branch" % sha1,
                    "%s even includes spaces" % sha2,
                )))
    repo = wrapper.repo
    assert read_gitlab_branches(repo) == {
        b'branch/some-branch': sha1.encode('ascii'),
        b'even includes spaces': sha2.encode('ascii'),
    }


def test_gitlab_branches_write_read(tmpdir):
    wrapper = RepoWrapper.init(tmpdir)
    repo = wrapper.repo

    branches = {b'topic/default/truc': b'01cafe34' * 5}

    write_gitlab_branches(repo, branches)
    assert read_gitlab_branches(repo) == branches


def test_gitlab_branches(tmpdir):
    wrapper = RepoWrapper.init(tmpdir)
    repo = wrapper.repo

    branches = {b'topic/default/truc': b'01cafe34' * 5}

    assert gitlab_branches(repo) is GITLAB_BRANCHES_MISSING

    write_gitlab_branches(repo, branches)
    # missing marker got invalidated
    assert gitlab_branches(repo) == branches
    # using cached value
    assert gitlab_branches(repo) == branches

    # values other than the missing marker get invalidated as well
    branches2 = {b'topic/default/thing': b'01cafe34' * 5}
    write_gitlab_branches(repo, branches2)
    assert gitlab_branches(repo) == branches2



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Heptapod specific branch facilities.

GitLab branches:

  GitLab branches are pointers to Mercurial changesets that the Rails
  application believes to be actual Git branches. They are updated
  automatically by each transaction, applying several rules tailored for
  good interaction with other GitLab components. They are not implemented
  as bookmarks because they are not meant to be exposed to Mercurial clients.

  GitLab branches are kept in a file that represents the last state that was
  exposed to other GitLab components (notably, the Rails application). In case
  there is an ongoing transaction, the file represents the state at the
  beginning of the transaction, and is updated at the end of transaction.
  Its uses include:

  - sending pre- and post-receive calls to the Rails application based on
    comparisons with the previous state.
  - fast querying by GitLab branch in HGitaly. While it is possible for HGitaly
    to implement the mapping to GitLab branches in a stateless way, it is very
    bad for performance to do it in all read operations.

  The state cannot be reconstructed, since it records what other components
  have already seen. But it can be reinitialized: GitLab will fail to update
  the Merge Requests that should have been affected by the current transaction,
  fail to create new Pipelines and generally miss all the consequences of
  the current transaction. Initializing it at import or in an otherwise
  empty transaction should not create any problem.

  File format: version preamble (common to all versions):

  - first 3 bytes: decimal string representation of the version number, padded
    with left zeros (spaces also accepted for reading)
  - byte 4: Unix End-Of-Line (LF, 0x10)

  File format: version 1 (after version preamble)

  - line-based, each line terminated the Unix way (LF, 0x10)
  - one GitLab branch per line
  - each line is made of the hexadecimal Mercurial changeset ID, followed by
    exactly one space character, then by the GitLab branch name (bytes with
    no encoding assumption). Example::

        01ca23fe01ca23fe01ca23fe01ca23fe01ca23fe branch/default

Default GitLab branch:
  This a GitLab branch, e.g., ``branch/default`` rather than a Mercurial
  branch, e.g., ``default``.

  It is stored inside the repository directory, but not in the main store.
  In principle, different shares could have different default GitLab branches,
  if that were to be useful.

  For comparison, GitLab uses the value of `HEAD` on Git repositories for
  this.
"""
from mercurial.i18n import _
from mercurial import (
    hg,
    error,
)

DEFAULT_GITLAB_BRANCH_FILE_NAME = b'default_gitlab_branch'
GITLAB_BRANCHES_FILE_NAME = b'gitlab.branches'
GITLAB_BRANCHES_FILE_CURRENT_VERSION = 1

GITLAB_BRANCHES_MISSING = object()
"""Cacheable marker to avoid repeated read attempts if missing"""


def read_gitlab_branches_file_version(fobj):
    """Read the version and leave `fobj` at the start of actual data.
    """
    version_line = fobj.read(4)
    # TODO proper exception
    assert version_line.endswith(b'\n')
    return int(version_line)  # ignores left 0 padding, whitespace and EOL


def write_gitlab_branches_file_version(
        fobj,
        version=GITLAB_BRANCHES_FILE_CURRENT_VERSION):
    fobj.write(b"%03d\n" % version)


def read_gitlab_branches(repo):
    """Return the GitLab branches mapping, as a dict.
    """
    # TODO exceptions other than `FileNotFoundException`, which is fine
    with repo.svfs(GITLAB_BRANCHES_FILE_NAME, mode=b'rb') as fobj:
        assert read_gitlab_branches_file_version(fobj) == 1
        return {branch.strip(): sha for sha, branch in
                (line.split(b' ', 1) for line in fobj)}


def gitlab_branches(repo):
    """Return the GitLab branches or GITLAB_BRANCHES_MISSING.

    The GitLab branches state file is still optional at this stage.
    Callers have to check for GITLAB_BRANCHES_MISSING and fall back to
    other means.

    Will eventually become a property on the repo object. Since our use case
    is different than what Mercurial usual filecaches are meant for, it
    is safer to do it directly in this first version.
    """
    glb = getattr(repo, '_gitlab_branches', None)
    if glb is not None:
        return glb

    try:
        glb = read_gitlab_branches(repo)
    except FileNotFoundError:
        glb = GITLAB_BRANCHES_MISSING
    repo._gitlab_branches = glb
    return glb


def write_gitlab_branches(repo, gl_branches):
    """Write the GitLab branches mapping atomically.
    """
    # cache invalidation. Done before the write so that we don't
    # get inconsistent results even in the unlikely case where
    # an exceptio would occur after file write actually happened.
    repo._gitlab_branches = None
    with repo.lock():
        with repo.svfs(GITLAB_BRANCHES_FILE_NAME,
                       mode=b'wb',
                       atomictemp=True,
                       checkambig=True) as fobj:
            write_gitlab_branches_file_version(fobj)
            for gl_branch, sha_hex in gl_branches.items():
                fobj.write(b"%s %s\n" % (sha_hex, gl_branch))
            # atomictempfile does not expose flush(), so there's no
            # point trying an fsync(). Mercurial does not seem to use
            # fsync() anyway.


def get_default_gitlab_branch(repo):
    """Return the default GitLab branch name, or ``None`` if not set."""
    branch = repo.vfs.tryread(DEFAULT_GITLAB_BRANCH_FILE_NAME)
    # (hg 5.4) tryread returns empty strings for missing files
    if not branch:
        return None
    return branch


def set_default_gitlab_branch(repo, target):
    if not target:
        raise error.Abort(_("The default GitLab branch cannot be an "
                            "empty string."))
    shared_from = hg.sharedreposource(repo)
    if shared_from is not None:
        repo = shared_from

    with repo.wlock():
        repo.vfs.write(DEFAULT_GITLAB_BRANCH_FILE_NAME, target)



# Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
from mercurial import (
    error,
)
import pytest

from ..testhelpers import RepoWrapper
from ..testhelpers import LocalRepoWrapper  # see test_compat_alis


parametrize = pytest.mark.parametrize


def test_write_commit_topic(tmpdir):
    """Demonstrate the use of write_commit with parent."""
    # it is essential to activate the rebase extension, even though
    # we don't use it in this test, because the
    # first loading of topic patches it only if it is present.
    # without this, all subsequent tests expecting rebase to preserve
    # topics would be broken
    wrapper = RepoWrapper.init(tmpdir,
                               config=dict(extensions=dict(rebase='',
                                                           topic='')))

    ctx0 = wrapper.write_commit('foo', content='Foo 0')
    wrapper.write_commit('foo', content='Foo 1')
    ctxbr = wrapper.write_commit('foo', content='Foo branch',
                                 parent=ctx0, topic='sometop')

    assert ctxbr.topic() == b'sometop'
    assert ctxbr.parents() == [ctx0]


def test_prune_update_hidden(tmpdir):
    wrapper = RepoWrapper.init(tmpdir,
                               config=dict(extensions=dict(evolve='')))
    wrapper.write_commit('foo', content='Foo 0')
    ctx = wrapper.write_commit('foo', content='Foo 1', return_ctx=True)
    wrapper.prune('.')
    assert ctx.obsolete()

    wrapper.update(0)
    assert tmpdir.join('foo').read() == 'Foo 0'

    with pytest.raises(error.FilteredRepoLookupError):
        wrapper.update(ctx.hex())

    wrapper.update_bin(ctx.node(), hidden=True)
    assert tmpdir.join('foo').read() == 'Foo 1'


def test_compat_alias(tmpdir):
    # that's enough to make flake8 happy and provides a strong
    # guarantee.
    assert RepoWrapper is LocalRepoWrapper



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
import pytest
from mercurial import (
    error,
)
from heptapod.testhelpers import (
    RepoWrapper,
)

from .utils import switch_user


def init_repo(basedir):
    return RepoWrapper.init(
        basedir,
        config=dict(
            hooks={'pretxnopen.check_write':
                   'python:heptapod.hooks.perm.check_write'
                   },
            web={'allow-push': 'developer',
                 },
        ))


def test_write_empty_repo(tmpdir):
    wrapper = init_repo(tmpdir)
    switch_user(wrapper, 'guest')
    with pytest.raises(error.HookAbort) as exc_info:
        wrapper.write_commit('foo', content='Foo', return_ctx=True)

    assert b'pretxnopen.check_write' in exc_info.value.args[0]
    switch_user(wrapper, 'developer')
    ctx = wrapper.write_commit('foo', message='developer is accepted',
                               return_ctx=True)
    assert ctx.description() == b'developer is accepted'


def test_phase_change(tmpdir):
    wrapper = init_repo(tmpdir)
    switch_user(wrapper, 'developer')
    ctx = wrapper.write_commit('foo', content='Foo', return_ctx=True)

    switch_user(wrapper, 'guest')
    with pytest.raises(error.HookAbort) as exc_info:
        wrapper.set_phase('public', [ctx.hex()])
    assert b'pretxnopen.check_write' in exc_info.value.args[0]


def test_wrong_hook(tmpdir):
    wrapper = init_repo(tmpdir)
    ui = wrapper.repo.ui
    pretxn = b'pretxnopen.check_write'
    hookdef = ui.config(b'hooks', pretxn)
    ui.setconfig(b'hooks', pretxn, b'')
    ui.setconfig(b'hooks', b'precommit.check_write', hookdef)
    # precommit because that one does not swallow exceptions other
    # than abort
    with pytest.raises(error.ProgrammingError) as exc_info:
        wrapper.write_commit('foo')

    assert 'precommit' in exc_info.value.args[0]



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Tests for the GitRefChange object.

(that may later be renamed as GitLabRefChange or other so that it doesn't
sound too weird once we are in full native mode)
"""
import pytest
from ...git import (
    GitRefChange as RefChange,  # `as` for potential rename
    ZERO_SHA,
)


class FakeHandler:

    def __init__(self, **git2hg):
        self.git2hg = git2hg

    def map_hg_get(self, git_sha):
        return self.git2hg.get(git_sha.decode())


def test_export_for_hook_errors():
    handler = FakeHandler(known_git_sha=b'123d12fa45')

    ch = RefChange(b'branch/something', b'known_git_sha', b'the_unknown')
    with pytest.raises(KeyError) as exc_info:
        ch.export_for_hook(handler=handler, native=True)
    assert exc_info.value.args[0] == b'the_unknown'

    ch = RefChange(b'branch/something', b'unknown', b'known_git_sha')
    assert ch.export_for_hook(handler=handler, native=True) == (
        ZERO_SHA, b'123d12fa45')



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Compatibility layer for externalisation of testhelpers."""

from mercurial_testhelpers.repo_wrapper import (  # noqa: F401
    RepoWrapper as CoreRepoWrapper,
    as_bytes,
    make_ui,
)


class RepoWrapper(CoreRepoWrapper):

    write_commit = CoreRepoWrapper.commit_file
    commit_remove = CoreRepoWrapper.commit_removal

    def prune(self, revs, successors=(), bookmarks=()):
        # the prune command expects to get all these arguments (it relies
        # on the CLI defaults but doesn't have any at the function call level).
        # They are unconditionally concatened to lists, hence must be lists.
        # (as of Mercurial 5.3.1)
        if isinstance(revs, (bytes, str)):
            revs = [revs]
        return self.command('prune',
                            rev=[as_bytes(r) for r in revs],
                            new=[],  # deprecated yet expected
                            # TODO py3 convert these two to bytes as needed:
                            successor=list(successors),
                            bookmark=list(bookmarks))

    def set_topic(self, topic):
        self.command('topics', as_bytes(topic))


RepoWrapper.register_commit_option('topic', 'set_topic')

# compatibility alias
LocalRepoWrapper = RepoWrapper



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Highest level tests for GitLab hook support.

These tests instantiate `Hook` and thus need a real repository.
Some use the API like in the hg-git wrapper, and thus need to monkey patch
a request recorder in.
"""
import io
import json
import pytest
import requests

from heptapod.testhelpers import (
    RepoWrapper,
    as_bytes,
)
from ..hooks import (
    Hook,
    PreReceive,
    PostReceive,
)
from .. import prune_reasons

ZERO_SHA = b"0" * 20

parametrize = pytest.mark.parametrize


def make_repo(base_dir, name):
    secret = base_dir.join('secret')
    secret.write("TEHTOKEN")
    wrapper = RepoWrapper.init(
        base_dir.join(name),
        config=dict(heptapod={'gitlab-shell': str(base_dir),
                              'gitlab-internal-api-secret-file': str(secret),
                              }),
    )
    # by default, ui.environ IS os.environ (shared) but that isn't true
    # in WSGI contexts. In that case, it is a plain dict, the WSGI environment
    # indeed. For these tests, we need them to be independent.
    wrapper.repo.ui.environ = {}
    return wrapper


def request_recorder(records, responses):
    """Record sent requests and treat each from the given responses.

    :param responses: a list of :class:`request.Response` instances or
                      :class:`Exception` instances. In the latter case,
                      the exception is raised instead of returning the
                      response.
    :param records: list to which the recorder appends the request that is
                    being sent.
    """
    def request(hook, meth, subpath, params, **kwargs):
        records.append((meth, subpath, params, kwargs))
        resp = responses.pop(0)
        if isinstance(resp, Exception):
            raise resp
        return resp

    return request


def make_response(status_code, body):
    """Create a :class:`request.Response` instance.

    :param body: must be `bytes`.
    """
    resp = requests.models.Response()
    resp.status_code = status_code
    resp.raw = io.BytesIO(body)
    return resp


def make_json_response(obj, status_code=200):
    return make_response(status_code, json.dumps(obj).encode())


def update_environ(repo, **kwargs):
    repo.ui.environ.update((as_bytes(k), as_bytes(v))
                           for k, v in kwargs.items())


def test_post_receive_params(tmpdir):
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = PostReceive(repo)

    update_environ(
        repo,
        HEPTAPOD_USERINFO_ID='23',
        GL_REPOSITORY='project-1024',
    )
    changes = {}, {
        b'branch/newbr': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
        b'topic/default/zz': (b"23cafe45" + b"0" * 12, b"1234" + b"b" * 16),
    }

    params = hook.params(changes)
    assert params['identifier'] == 'user-23'
    assert set(params['changes'].splitlines()) == {
        b'23cafe45000000000000 1234bbbbbbbbbbbbbbbb topic/default/zz',
        b'00000000000000000000 beef1234deadaaaaaaaa branch/newbr'}
    assert params['gl_repository'] == 'project-1024'


@parametrize('with_mr', ('with-mr', 'no-mr'))
def test_pre_receive_params(tmpdir, with_mr):
    # string value was neat for pytest output, prefering a bool from now on
    with_mr = with_mr == 'with-mr'
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = PreReceive(repo)

    update_environ(
        repo,
        HEPTAPOD_USERINFO_ID='2',
        GL_REPOSITORY='project-3145926',
    )
    if with_mr:
        update_environ(repo, HEPTAPOD_ACCEPT_MR_IID='123')

    changes = 'prune-reasons-are-ignored', {
        b'topic/default/with-a-dash': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
        b'branch/default': (b"23cafe45" + b"0" * 12, b"1234" + b"b" * 16),
    }
    params = hook.params(changes)

    assert params['user_id'] == '2'
    assert set(params['changes'].splitlines()) == {
        b'23cafe45000000000000 1234bbbbbbbbbbbbbbbb branch/default',
        b'00000000000000000000 beef1234deadaaaaaaaa topic/default/with-a-dash'}
    assert params['gl_repository'] == 'project-3145926'
    assert params['project'].encode() == repo.root
    if with_mr:
        assert params['accept_mr_iid'] == 123
    else:
        assert 'accept_mr_iid' not in params


def test_post_receive_params_prune(tmpdir):
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = PostReceive(repo)

    update_environ(
        repo,
        HEPTAPOD_USERINFO_ID='2',
        GL_REPOSITORY='project-3145926',
    )
    pruned = {
        b'topic/default/mytop': prune_reasons.TopicPublished(b'cafe5678'),
        b'master': prune_reasons.BookmarkRemoved(b'1234'),
        b'branch/other': prune_reasons.BranchClosed([
            (b'239acf', [b'968de76'])
        ])
    }
    git_changes = {
        b'topic/default/with-a-dash': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
        b'branch/default': (b"23cafe45" + b"0" * 12, b"1234" + b"b" * 16),
    }
    params = hook.params((pruned, git_changes))

    assert params['identifier'] == 'user-2'
    assert set(params['changes'].splitlines()) == {
        b'23cafe45000000000000 1234bbbbbbbbbbbbbbbb branch/default',
        b'00000000000000000000 beef1234deadaaaaaaaa topic/default/with-a-dash'}
    assert params['gl_repository'] == 'project-3145926'

    # GitLab branches are serialized in base64, since have unknown encoding
    assert json.loads(params['hg_prunes']) == {
        'topic_published': {'dG9waWMvZGVmYXVsdC9teXRvcA==': 'cafe5678'},
        'bookmark_removed': {'bWFzdGVy': '1234'},
        'branch_closed': {'YnJhbmNoL290aGVy': [['239acf', ['968de76']]]},
    }


def test_post_receive_handle_response_gl_12_2(tmpdir):
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = PostReceive(repo)

    out = hook.handle_response(
        200,
        # example taken from the functional tests
        {u'broadcast_message': None,
         u'reference_counter_decreased': True,
         u'merge_request_urls': [
             {u'url': u'http://localhost:3000/root/'
              'test_project_1583330760_9216309/merge_requests/1',
              u'branch_name': u'topic/default/antelope',
              u'new_merge_request': False}
         ]})[1]
    out = [l.strip() for l in out.splitlines() if l.strip()]
    assert out == [
        b'View merge request for topic/default/antelope:',
        b'http://localhost:3000/root/test_project_1583330760_9216309/'
        b'merge_requests/1']

    broadcast = "Check out the new tutorial"
    out = hook.handle_response(
        200,
        {u'broadcast_message': broadcast,
         u'reference_counter_decreased': True,
         })[1]
    out = (l.strip() for l in out.splitlines() if l.strip())
    assert any(broadcast.encode() in l for l in out)

    project_created = "Congrats!"
    out = hook.handle_response(
        200,
        {u'project_created_message': project_created,
         u'reference_counter_decreased': True,
         })[1]
    out = (l.strip() for l in out.splitlines() if l.strip())
    assert any(project_created.encode() in l for l in out)

    # GitLab internal API gives warnings as a unique string
    # see lib/api/helpers/internal_helpers.rb
    warnings = "You probably shouldn't have"
    out = hook.handle_response(
        200,
        {u'warnings': warnings,
         u'reference_counter_decreased': True,
         })[1]
    out = (l.strip() for l in out.splitlines() if l.strip())
    assert any(warnings.encode() in l for l in out)


def test_post_receive_handle_response_gl_12_3(tmpdir):
    wrapper = make_repo(tmpdir, 'proj.hg')
    repo = wrapper.repo
    hook = PostReceive(repo)

    out = hook.handle_response(
        200,
        # example taken from the functional tests
        {u'messages': [
            {u'message': (
                u'To create a merge request '
                u'for topic/default/ant, visit:\n'
                u'http://localhost:3000/root/test_project/merge_requests/new'
                u'?merge_request%5Bsource_branch%5D=topic%2Fdefault%2Fant'),
             u'type': u'basic',
             },
            {u'message': u'This is a TESTING SERVER',
             u'type': u'alert',
             },
        ]})[1]
    out = [l.strip() for l in out.splitlines() if l.strip()]
    assert out == [
        b'To create a merge request for topic/default/ant, visit:',
        b'http://localhost:3000/root/test_project/merge_requests/new'
        b'?merge_request%5Bsource_branch%5D=topic%2Fdefault%2Fant',
        b'=' * 72,
        b'This is a TESTING SERVER',
        b'=' * 72,
    ]


def test_integration_post_receive(tmpdir, monkeypatch):
    records = []
    responses = [
        make_json_response(
            {u'broadcast_message': None,
             u'reference_counter_decreased': True,
             u'merge_request_urls': [
                 {u'url': u'http://localhost:3000/root/'
                  'test_project_1583330760_9216309/merge_requests/1',
                  u'branch_name': u'topic/default/antelope',
                  u'new_merge_request': False}
             ]}),
        requests.exceptions.Timeout("1 ns is too much ;-)"),
    ]

    monkeypatch.setattr(Hook, 'gitlab_internal_api_request',
                        request_recorder(records, responses))

    repo_wrapper = make_repo(tmpdir, 'repo.hg')
    repo = repo_wrapper.repo

    update_environ(
        repo,
        HEPTAPOD_USERINFO_ID='23',
        GL_REPOSITORY='project-1024',
    )

    hook = PostReceive(repo_wrapper.repo)
    changes = {}, {
        b'branch/newbr': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
        b'topic/default/zz': (b"23cafe45" + b"0" * 12, b"1234" + b"b" * 16),
    }

    return_code, out, err = hook(changes)
    assert return_code == 0
    out = [l.strip() for l in out.splitlines() if l.strip()]
    assert out == [
        b'View merge request for topic/default/antelope:',
        b'http://localhost:3000/root/test_project_1583330760_9216309/'
        b'merge_requests/1']

    return_code, out, err = hook(changes)
    assert return_code == 1
    assert err == 'GitLab is unreachable'


def test_integration_pre_receive(tmpdir, monkeypatch):
    records = []
    responses = [make_json_response({u'message': "yes, granted!",
                                     u'status': True,
                                     },
                                    status_code=200),
                 make_json_response({u'status': True},
                                    status_code=200),
                 make_json_response({u'message': "go away!",
                                     u'status': False,
                                     },
                                    status_code=401),
                 make_response(504, b"Bad gateway :-("),
                 requests.exceptions.ConnectionError("Failed"),
                 ]
    monkeypatch.setattr(Hook, 'gitlab_internal_api_request',
                        request_recorder(records, responses))

    repo_wrapper = make_repo(tmpdir, 'repo.hg')
    repo = repo_wrapper.repo

    update_environ(
        repo,
        HEPTAPOD_USERINFO_ID='23',
        GL_REPOSITORY='project-1024',
    )

    hook = PreReceive(repo_wrapper.repo)
    changes = {}, {
        b'branch/newbr': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
        b'topic/default/zz': (b"23cafe45" + b"0" * 12, b"1234" + b"b" * 16),
    }

    return_code, out, err = hook(changes)
    assert return_code == 0
    assert out == b'yes, granted!'

    return_code, out, err = hook(changes)
    assert return_code == 0
    assert out == b''

    return_code, out, err = hook(changes)
    assert return_code == 1
    assert err == 'go away!'

    return_code, out, err = hook(changes)
    assert return_code == 1
    assert err == 'API is not accessible'

    return_code, out, err = hook(changes)
    assert return_code == 1
    assert err == 'GitLab is unreachable'


def test_empty_changes(tmpdir):
    hook = PostReceive(make_repo(tmpdir, 'repo.hg').repo)
    code, out, err = hook(())
    assert code == 0
    assert not out
    assert not err


def test_str_bytes(tmpdir):
    # these are used in various messages
    hook = PostReceive(make_repo(tmpdir, 'repo.hg').repo)
    assert b'post-receive' in hook.__bytes__()
    assert 'post-receive' in str(hook)


def test_missing_secret(tmpdir):
    repo = RepoWrapper.init(tmpdir).repo
    shell = tmpdir.join('heptapod-shell').ensure(dir=True)
    repo.ui.setconfig(b'heptapod', b'gitlab-shell', str(shell).encode())
    with pytest.raises(RuntimeError) as exc_info:
        PostReceive(repo)
    assert "secret-file" in exc_info.value.args[0]

    # defaulting to gitlab-shell secret if available
    shell.join('.gitlab_shell_secret').write('myseekret\n')
    hook = PostReceive(repo)
    assert hook.gitlab_secret == 'myseekret'


def test_hook_missing_user(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    hook = PreReceive(repo)

    repo.ui.environ[b'GL_REPOSITORY'] = b'project-1024'
    with pytest.raises(ValueError) as exc_info:
        hook.environ()
    assert 'User id' in exc_info.value.args[0]

    # and that gets catched, does not break the process
    code, out, err = hook(dict(master=(0, 1)))
    assert code != 0


def test_hook_missing_project(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    hook = PostReceive(repo)

    repo.ui.environ[b'HEPTAPOD_USERINFO_ID'] = b'3'
    with pytest.raises(ValueError) as exc_info:
        hook.environ()
    assert 'GL_REPOSITORY' in exc_info.value.args[0]

    # and that gets catched, does not break the process
    code, out, err = hook(dict(master=(0, 1)))
    assert code != 0


def test_hook_skip(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    repo.ui.environ[b'HEPTAPOD_SKIP_GITLAB_HOOK'] = b'pre-receive'
    hook = PreReceive(repo)
    assert hook(dict(master=(0, 1))) == (0, b'', '')

    repo.ui.environ[b'HEPTAPOD_SKIP_GITLAB_HOOK'] = b'post-receive'
    hook = PostReceive(repo)
    assert hook(dict(master=(0, 1))) == (0, b'', '')


def test_hook_skip_all(tmpdir):
    wrapper = make_repo(tmpdir, 'proj2.hg')
    repo = wrapper.repo
    repo.ui.environ[b'HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'] = b'no'
    hook = PreReceive(repo)
    # it's been called and failed because of lack of User id
    assert hook(dict(master=(0, 1)))[0] == 1

    repo.ui.environ[b'HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'] = b'yes'
    hook = PreReceive(repo)
    assert hook(dict(master=(0, 1))) == (0, b'', '')
